#
# Setup environment variable for various directories.
#
# Copyright (C) 2008-2012 Intel Corp.
#   Author: Huang Ying <ying.huang@intel.com>
#
# This file is released under the GPLv2.
#

SDIR=$(script_dir)
CDIR=$ROOT/cases/coverage/soft-inj
CONF_DIR=$SDIR/config
RDIR=$LOG_DIR/$driver
WDIR=$TMP_DIR/$driver
LDIR=$ROOT/lib
TDIR=$TOOL_DIR
